/* Copyright (C) 2015-2021 |Meso|Star> (contact@meso-star.com)
 *
 * This software is a computer program whose purpose is to describe a
 * virtual 3D environment that can be ray-traced and sampled both robustly
 * and efficiently.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "s3d_device_c.h"
#include "s3d_sphere.h"

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static void
sphere_release(ref_T* ref)
{
  struct sphere* sphre;
  struct s3d_device* dev;
  ASSERT(ref);

  sphre = CONTAINER_OF(ref, struct sphere, ref);
  dev = sphre->dev;
  MEM_RM(dev->allocator, sphre);
  S3D(device_ref_put(dev));
}

/*******************************************************************************
 * Local functions
 ******************************************************************************/
res_T
sphere_create(struct s3d_device* dev, struct sphere** out_sphere)
{
  struct sphere* sphere = NULL;
  res_T res = RES_OK;
  ASSERT(dev && out_sphere);

  sphere = (struct sphere*)MEM_CALLOC(dev->allocator, 1, sizeof(struct sphere));
  if(!sphere) {
    res = RES_MEM_ERR;
    goto error;
  }
  ref_init(&sphere->ref);
  S3D(device_ref_get(dev));
  sphere->dev = dev;
  sphere->radius = -1;

exit:
  *out_sphere = sphere;
  return res;
error:
  if(sphere) {
    sphere_ref_put(sphere);
    sphere = NULL;
  }
  goto exit;
}

void
sphere_ref_get(struct sphere* sphere)
{
  ASSERT(sphere);
  ref_get(&sphere->ref);
}

void
sphere_ref_put(struct sphere* sphere)
{
  ASSERT(sphere);
  ref_put(&sphere->ref, sphere_release);
}

